"""

    Package Exceptions
    ------------------

"""


class BiometricsClientError(Exception):
    """Base biometrics client error"""

    pass


class BiometricsApiError(BiometricsClientError):
    """Generic biometrics api request error."""

    pass


class BiometricsApiRequestError(BiometricsClientError):
    """Generic biometrics api error."""

    pass


class BiometricsApiResultsNotReadyError(BiometricsClientError):
    """Results not ready error."""

    pass


class BiometricsApiInternalError(BiometricsClientError):
    """Internal Biometrics API Error."""

    pass



"""

    API Client
    ----------

"""
import requests
from biometrics_client import __version__
from biometrics_client.exceptions import (
    BiometricsApiError,
    BiometricsApiRequestError,
    BiometricsApiInternalError,
    BiometricsApiResultsNotReadyError,
)
from biometrics_client._utils import (
    create_multipart_encoder,
    format_error_message,
    not_ready_signal,
    task_waiter,
)
from pathlib import Path
from urllib.parse import urljoin
from requests.models import Response
from typing import Any, Dict, List, Union, Tuple, Optional, cast


_DEFAULT_HEADERS = {"User-Agent": f"Biometrics-Client v{__version__}"}


class ElementHumanBiometrics:
    """Client for the Element Human Biometrics API.

    Args:
        access_key (str): an access key for the API.
        secret_key (str): a secret key for the API.
        timeout (int): the maximum amount of time to wait for a
            response from the server in seconds.
        url (str): the URL for the API
        verbose (bool): if True print additional information.

    """

    def __init__(
        self,
        access_key: str,
        secret_key: str,
        timeout: int = 30,
        url: str = "https://biometrics.elementapis.com/public/v0.1/",
        verbose: bool = True,
    ) -> None:
        if not isinstance(url, str) or not url:
            raise TypeError("`url` must be a non-null string.")
        elif not isinstance(access_key, str):
            raise TypeError(f"`access_key` not of type str, got {access_key}.")
        elif not isinstance(secret_key, str):
            raise TypeError(f"`secret_key` not of type str, got {secret_key}.")
        self._access_key = access_key
        self._secret_key = secret_key

        self.timeout = timeout
        self.url = url
        self.verbose = verbose

    @property
    def _credentials(self) -> Dict[str, str]:
        """API Credentials"""
        return {"x-access-key": self._access_key, "x-secret-key": self._secret_key}

    def _print(self, msg: str) -> None:
        if self.verbose:
            print(msg)

    @staticmethod
    def _response_validator(r: Response) -> None:
        try:
            r.raise_for_status()
        except requests.exceptions.RequestException:
            msg = (
                f"Bad response from biometrics api, "
                f"got message: {format_error_message(r).rstrip('.')}."
            )
            if 400 <= r.status_code < 500:
                raise BiometricsApiRequestError(msg)
            elif 500 <= r.status_code < 600:
                raise BiometricsApiInternalError(msg)
            else:
                raise BiometricsApiError(msg)

    def ping(self, **kwargs: Any) -> Dict[str, str]:
        """Ping the API.

        Args:
            **kwargs (Keyword Args): Keyword arguments to pass to
            ``requests``.

        Returns:
            dict

        """
        r = requests.get(urljoin(self.url, "ping"), **kwargs)
        self._response_validator(r)
        return cast(dict, r.json())

    def apply(
        self,
        video_file_path: Path,
        metadata_file_path: Optional[Path] = None,
        analyses: Union[str, List[str], Tuple[str, ...]] = ("emotion",),
        **kwargs: Any,
    ) -> Dict[str, Union[str, Dict[str, str]]]:
        """Send a video to the Biometrics API for analysis

        Args:
            video_file_path (Path): a system path to a video
            metadata_file_path (Path, optional): a path to a metadata file.
            analyses (str, list, tuple): analyses to perform on the video.

                * if a string, must be 'all'
                * if a list of strings or a tuple of strings
                    defining analyses to perform. These can be any of
                    the following:

                      * 'face': Face bound box.
                      * 'eye': Eye bounding boxes. Depends on: 'face'.
                      * 'emotion': compute Ekman emotions for the video,
                                   along with quality metrics. Depends on: 'face'.
                      * 'gaze': eye gaze. Depends on: 'face', 'eye'.

            **kwargs (Keyword Args): Keyword arguments to pass to
                ``requests``.

        Returns:
            response (dict)

        """
        multipart_data = create_multipart_encoder(
            video_file_path, metadata_file_path=metadata_file_path
        )
        r = requests.post(
            urljoin(self.url, "apply"),
            data=multipart_data,
            timeout=self.timeout,
            params=dict(analyses=analyses),
            headers={
                "Content-Type": multipart_data.content_type,
                **self._credentials,
                **_DEFAULT_HEADERS,
            },
            **kwargs,
        )
        self._response_validator(r)
        return cast(dict, r.json())

    def results(
        self,
        task_id: str,
        check_interval: int = 10,
        max_wait: Optional[int] = None,
        **kwargs: Any,
    ) -> Dict[str, Any]:
        """Get a task from the Biometrics API.

        Args:
            task_id (str): a task ID obtained from the /apply endpoint
            check_interval (int): time to wait between checks to determine
                if results are ready. Only applies if ``max_wait`` is not ``None``.
            max_wait (int, optional): the maximum amount of time to wait
                for the results in seconds.
            **kwargs (Keyword Args): Keyword arguments to pass to
                ``requests``.

        Returns:
            response (dict)

        Raises:
            BiometricsApiResultsNotReadyError: if results are not yet ready.

        """

        def fetch() -> Dict[str, Any]:
            r = requests.get(
                urljoin(self.url, f"results/{task_id}"),
                timeout=self.timeout,
                headers={**self._credentials, **_DEFAULT_HEADERS},
                **kwargs,
            )
            if not_ready_signal(r):
                raise BiometricsApiResultsNotReadyError(format_error_message(r))
            self._response_validator(r)
            return cast(dict, r.json())

        output = task_waiter(
            func=fetch,
            max_wait=max_wait,
            sleep_time=check_interval,
            handled_exceptions=(BiometricsApiResultsNotReadyError,),
            timeout_exception=requests.exceptions.ConnectTimeout(
                f"Timed out waiting for task '{task_id}'"
            ),
        )
        return cast(dict, output)

    def apply_and_wait(
        self,
        video_file_path: Path,
        metadata_file_path: Optional[Path] = None,
        analyses: Union[str, List[str], Tuple[str, ...]] = ("emotion",),
        max_wait: Optional[int] = 60 * 30,
        **kwargs: Any,
    ) -> Tuple[str, Dict[str, Any]]:
        """Send a video to the Biometrics API for analysis
        and wait for the results.

        Args:
            video_file_path (Path): a system path to a video
            metadata_file_path (Path, optional): a path to a metadata file.
            analyses (str, list, tuple): analyses to perform on the video.

                * if a string, must be 'all'
                * if a list of strings or a tuple of strings
                    defining analyses to perform. These can be any of
                    the following:

                      * 'face': Face bound box.
                      * 'eye': Eye bounding boxes. Depends on: 'face'.
                      * 'emotion': compute Ekman emotions for the video,
                                   along with quality metrics. Depends on: 'face'.
                      * 'gaze': eye gaze. Depends on: 'face', 'eye'.

            max_wait (int, optional): the maximum amount of time to wait
                for the results in seconds.
            **kwargs (Keyword Args): Keyword arguments to pass to
                ``requests``.

        Returns:
            tuple:
                task_id (str): the task id
                response (dict): the response payload.

        Raises:
            BiometricsApiResultsNotReadyError: if results are not yet ready.

        """
        task = self.apply(
            video_file_path=video_file_path,
            metadata_file_path=metadata_file_path,
            analyses=analyses,
            **kwargs,
        )
        task_id = task["response"]["task_id"]  # type: ignore
        self._print(f"Upload Complete. Task ID: {task_id}.")
        return task_id, self.results(task_id, max_wait=max_wait, **kwargs)



#!/usr/bin/env python3
"""

    Element setup
    ~~~~~~~~~~~~~

"""
import sys
import shlex
import subprocess
from pathlib import Path
from setuptools import setup, find_packages  # type: ignore
from setuptools.command.develop import develop  # type: ignore
from typing import List


def read_readme() -> str:
    readme_file_path = Path(__file__).resolve().with_name("README.md")
    return readme_file_path.read_text("utf-8")


def parse_reqs(requirements_file: str) -> List[str]:
    """Get requirements as a list of strings from the file."""
    with open(requirements_file) as f:
        return [i.rstrip() for i in f.readlines()]


class CustomDevelop(develop):
    """Develop command that actually prepares the development environment."""

    def run(self) -> None:
        """Setup the local dev environment fully."""
        super().run()

        for command in [
            "pip --version",
            "pip install -r dev_requirements.txt",
            "pip install -r requirements.txt",
        ]:
            print("\nCustom develop - executing:", command, file=sys.stderr)
            subprocess.check_call(shlex.split(command))


setup(
    name="eh-biometrics-client",
    version="0.0.1",
    url="https://github.com/elementhuman/biometrics-client",
    author="Element Human",
    description=read_readme(),
    packages=find_packages(exclude=["tests"]),
    install_requires=parse_reqs("requirements.txt"),
    extras_require={"test": ["pytest", "coverage"]},
    cmdclass={"develop": CustomDevelop},
)



"""

    Utils
    ~~~~~

"""
import io
import time
import logging
from pathlib import Path
from os.path import basename
from requests import Response  # type: ignore
from requests_toolbelt import MultipartEncoder  # type: ignore
from typing import Any, Callable, Type, Tuple, Optional

log = logging.getLogger(__name__)


def _open_as_bytes(path: Path) -> io.BytesIO:
    with path.open("rb") as f:
        return io.BytesIO(f.read())


def _get_file_type(path: Path) -> str:
    return path.suffix.lstrip(".")


def task_waiter(
    func: Callable[..., Any],
    max_wait: Optional[int],
    sleep_time: int,
    handled_exceptions: Optional[Tuple[Type[BaseException], ...]] = None,
    timeout_exception: Optional[BaseException] = None,
) -> Any:
    """Wait for ``func`` to successfully return.

    Args:
        func (callable): a function to run
        max_wait (int, optional): the maximum amount of time to
            wait for ``func`` to successfully return.
        sleep_time (int): the amount of time to sleep between
            executions of ``func``
        handled_exceptions (tuple, optional): one or
            more exceptions to handle
        timeout_exception (BaseException, optional): an
            exception to raise if ``max_wait`` is reached.

    Returns:
        Any

    """
    if max_wait is None:
        return func()

    start_time = time.time()
    while (time.time() - start_time) < max_wait:
        try:
            return func()
        except handled_exceptions or tuple():
            time.sleep(sleep_time)
    else:
        if timeout_exception is not None:
            raise timeout_exception
        return None


def not_ready_signal(r: Response) -> bool:
    """Return or not ``r`` is ready.

    Args:
        r (Response): response object.

    Returns:
        bool

    """
    return r.status_code == 400 and "not ready" in r.text.lower()


def format_error_message(r: Response) -> str:
    """Format an error message.

    Args:
        r (Response): a response object

    Returns:
        str

    """
    try:
        payload = r.json()
        if isinstance(payload, dict) and set(payload.keys()) == {"message"}:
            return str(payload["message"])
        else:
            return str(r.text)
    except BaseException:
        log.exception("Message extraction failed, got %s.", r.text)
        return str(r.text)


def create_multipart_encoder(
    video_file_path: Path, metadata_file_path: Optional[Path] = None
) -> MultipartEncoder:
    """Create a multipart encoder

    Args:
        video_file_path (Path): a system path to a video
        metadata_file_path (Path, optional): a path to a metadata file.

    Returns:
        MultipartEncoder

    """

    def make_value(path: Path) -> Tuple[str, io.BytesIO, str]:
        return (
            basename(str(path)),
            _open_as_bytes(path),
            f"video/{_get_file_type(path)}",
        )

    fields = dict(video_file=make_value(video_file_path))
    if metadata_file_path:
        fields["metadata"] = make_value(metadata_file_path)
    return MultipartEncoder(fields)
